#if 0
hh(00...1f) ll(00...ff)
00           data end
00+hh ...    short-raw-data  (hh) bytes raw data
20+hh ll ... long-raw-data (hhll) bytes raw data
40+hh ll     fill 4 + (hh) bytes with (ll)
60+hh        copy more (hh) bytes
E0+hh ll     copy 7 bytes from current output address - (hhll)
80...DF      unknown
#endif

#include <stdio.h>
#define MUS_MAX 0x1000

enum {
ERROR_INPUT = -64,
ERROR_OUTPUT,
ERROR_BACKSTEP,
ERROR_SIZE,
ERROR_ENDMARK,
ERROR_UNKNOWN1,
ERROR_UNKNOWN2
};

const char * const errstr[] = {
	"input data too small",
	"output data too large",
	"over backstep limit",
	"input data remain",
	"missing $00",
	"$6x before $Ex",
	"unknown code $80-$DF",
};

int decode_ds8music(unsigned char *outbuf, int ol, unsigned char *inbuf, int il)
{
#define OUTPUT(x) { if (op == ol) return ERROR_OUTPUT; outbuf[op] = x; op++; }
	int ip = 2, op = 0, hh,mm,ll,l,w = -1;

	l = (inbuf[1] << 8) + inbuf[0];
	if (il - l < 1) return ERROR_INPUT;

	il = l;
	if (inbuf[il] != 0) return ERROR_ENDMARK;

	while (ip < il)
	{
		switch (inbuf[ip] & 0xe0)
		{
			case 0x00:
				hh = inbuf[ip++] & 0x1f;
				l = hh;
				if (il - ip < l) return ERROR_INPUT;
				while (l--)
				{
					OUTPUT(inbuf[ip]);
					ip++;
				}
				break;
			case 0x20:
				hh = inbuf[ip++] & 0x1f;
				if (il - ip < 1) return ERROR_INPUT;
				mm = inbuf[ip++] & 0xff;
				l = (hh << 8) + mm;
				if (il - ip < l) return ERROR_INPUT;
				while (l--)
				{
					OUTPUT(inbuf[ip]);
					ip++;
				}
				break;
			case 0x40:
				hh = inbuf[ip++] & 0x1f;
				if (il - ip < 1) return ERROR_INPUT;
				mm = inbuf[ip++] & 0xff;
				l = hh + 4;
				while (l--) OUTPUT(mm);
				break;
			case 0x60:
				hh = inbuf[ip++] & 0x1f;
				l = hh;
				if (w == -1) return ERROR_UNKNOWN1;
				if (op < w) return ERROR_BACKSTEP;
				while (l--) OUTPUT(outbuf[op - w]);
				break;
			case 0xE0:
				hh = inbuf[ip++] & 0x1f;
				if (il - ip < 1) return ERROR_INPUT;
				mm = inbuf[ip++] & 0xff;
				w = (hh << 8) + mm;
				l = 7;
				if (op < w) return ERROR_BACKSTEP;
				while (l--) OUTPUT(outbuf[op - w]);
				break;
			default:
				return ERROR_UNKNOWN2;
		}
	}
	if (ip != il) return ERROR_SIZE;
	return op;
}


unsigned char ds8_hed[0x200] = {
	0x48, 0x45, 0x53, 0x4D, 0x00, 0x00, 0x40, 0xE0, 
	0xFF, 0xF8, 0x01, 0x02, 0x03, 0x04, 0xF9, 0x00, 
	0x44, 0x41, 0x54, 0x41, 0xE0, 0x3F, 0x00, 0x00, 
	0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x4C, 0x00, 0xE1, 0x4C, 0x40, 0xE1, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x78, 0x48, 0xD4, 0x20, 0x80, 0xE0, 0x68, 0x20, 
	0xC0, 0xE0, 0xA9, 0xFD, 0x8D, 0x02, 0x14, 0xA9, 
	0x05, 0x8D, 0x00, 0x00, 0xA9, 0xC8, 0x8D, 0x02, 
	0x00, 0x58, 0x60, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x72, 0x65, 0x73, 0x65, 0x74, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x73, 0x00, 0xE5, 0x00, 0xC0, 0x00, 0x10, 0xA9, 
	0x80, 0x85, 0x4B, 0xA9, 0xFF, 0x8D, 0x01, 0x08, 
	0x60, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x72, 0x65, 0x71, 0x75, 0x65, 0x73, 0x74, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0xAA, 0xBD, 0x00, 0xE2, 0x8D, 0xF1, 0xCF, 0xBD, 
	0x00, 0xE3, 0x8D, 0xF8, 0xCF, 0xBD, 0x00, 0xE4, 
	0x8D, 0xF9, 0xCF, 0x20, 0xF0, 0xCF, 0x20, 0x51, 
	0xC6, 0x20, 0x84, 0xC0, 0x4C, 0x12, 0xC0, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x70, 0x6C, 0x61, 0x79, 0x28, 0x76, 0x73, 0x79, 
	0x6E, 0x63, 0x29, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x5A, 0xDA, 0x48, 0xAD, 0x00, 0x00, 0x20, 0x43, 
	0xC1, 0x68, 0xFA, 0x7A, 0x40, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x70, 0x63, 0x6D, 0x28, 0x74, 0x69, 0x6D, 0x65, 
	0x72, 0x29, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x4C, 0xDE, 0xF4, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x5A, 0xDA, 0x48, 0x8D, 0x03, 0x14, 0xEA, 0xEA, 
	0xEA, 0x68, 0xFA, 0x7A, 0x40, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x48, 0xB5, 0x00, 0x65, 0x30, 0x95, 0x00, 0xE8, 
	0xB5, 0x00, 0x65, 0x31, 0x95, 0x00, 0x68, 0x02, 
	0x60, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x08, 0x48, 0xBD, 0xE0, 0xCA, 0xFA, 0x28, 0x60, 
};

unsigned char player_bin[0x100] = {
	0x78, 0xAD, 0x08, 0xE0, 0x53, 0x01, 0xAD, 0x09, 
	0xE0, 0x53, 0x02, 0xAD, 0x0A, 0xE0, 0x53, 0x04, 
	0xAD, 0x0B, 0xE0, 0x53, 0x08, 0xAD, 0x0C, 0xE0, 
	0x53, 0x10, 0xAD, 0x0D, 0xE0, 0x53, 0x20, 0xAD, 
	0x0E, 0xE0, 0x53, 0x40, 0xAD, 0x05, 0xE0, 0x8D, 
	0xFF, 0x3F, 0x9C, 0xFE, 0x3F, 0x20, 0xF0, 0xFF, 
	0x4C, 0x40, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0xA9, 0x01, 0x8D, 0x00, 0x10, 0xA9, 0x03, 0x8D, 
	0x00, 0x10, 0x82, 0xA9, 0x01, 0x8D, 0x00, 0x10, 
	0x48, 0x68, 0xEA, 0xBD, 0x80, 0x3F, 0x9D, 0x80, 
	0x3F, 0xAD, 0x00, 0x10, 0x9C, 0x00, 0x10, 0x0A, 
	0x0A, 0x0A, 0x0A, 0x8D, 0xFD, 0x3F, 0xAD, 0x00, 
	0x10, 0x29, 0x0F, 0x0D, 0xFD, 0x3F, 0x49, 0xFF, 
	0x8D, 0xFD, 0x3F, 0x4D, 0xFE, 0x3F, 0x2D, 0xFD, 
	0x3F, 0x8D, 0xFC, 0x3F, 0xAD, 0xFD, 0x3F, 0x8D, 
	0xFE, 0x3F, 0x93, 0x01, 0xFC, 0x3F, 0xF0, 0x0A, 
	0xAD, 0xFF, 0x3F, 0x1A, 0x8D, 0xFF, 0x3F, 0x20, 
	0xF0, 0xFF, 0x93, 0x02, 0xFC, 0x3F, 0xF0, 0x0A, 
	0xAD, 0xFF, 0x3F, 0x3A, 0x8D, 0xFF, 0x3F, 0x20, 
	0xF0, 0xFF, 0x4C, 0x40, 0xFF, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x6C, 0x06, 0xE0, 0x00, 0x40, 0x00, 0xF4, 0xFF, 
	0x20, 0xE0, 0x23, 0xE0, 0xF4, 0xFF, 0x00, 0xFF, 
};

/*
 0000-01ff header
 0200-02ff table bank
 0300-03ff table lo
 0400-04ff table hi
 0500-14ff driver
 1f00-1fff player.bin
 2000-     music
*/
#define DRV_TOP 0x0500

int searchsong(FILE *ifp , long *pip)
{
	int c;
	fseek(ifp, (*pip) + 2, SEEK_SET);
	for ( ;!feof(ifp) && !ferror(ifp); (*pip)++)
	{
		unsigned char buf[14];
		c = fgetc(ifp);
		if (c != 0x0c) continue;
		if (fread(buf,1,14,ifp) != 14) break;
		fseek(ifp, (*pip) + 3, SEEK_SET);
		if (buf[0] == 0 || (buf[0] & 0xf) != 0) continue;
/*
		if (buf[0] != 0x20 && buf[0] != 0x40 && buf[0] != 0x60 && buf[0] != 0x80) continue;
*/
		if (buf[1] != 0x00) continue;
#if 1
		if (buf[12] != 0x4c && buf[12] != 0x4d) continue;
		if (buf[13] != 0x00) continue;
#endif
		return 1;
	}
	return 0;
}

int main(int argc, char **argv)
{
	char *iname, *oname = "ds8.hes";
	unsigned char header[0x2000];
	long ip, op;
	int il, song, b=0,h=0,l=0;

	FILE *ifp, *ofp;
	if (argc < 2 | 3 < argc)
	{
		fprintf(stderr, "usage:\n");
		fprintf(stderr, "\tDS8HES image-filename [hes-filename]\n");
		return 1;
	}
	iname = argv[1];
	if (argc == 3) oname = argv[2];
	ifp = fopen(iname, "rb");
	if (!ifp)
	{
		fprintf(stderr, "read error: %s\n", iname);
		return 1;
	}
	ofp = fopen(oname, "wb");
	if (!ofp)
	{
		fprintf(stderr, "write error: %s\n", oname);
		return 1;
	}


	memcpy(&header[0x0000], ds8_hed, 0x200);
	memset(&header[0x0200], 0, 0x1D00);
	memcpy(&header[0x1f00], player_bin, 0x100);

	fseek(ifp, 0x4800, SEEK_SET);
	fread(&header[DRV_TOP], 1, 0x1000, ifp);

	/* patch driver */
	header[DRV_TOP + 0x0645] = 0x60;

	header[DRV_TOP + 0x042c + 0] = 0x4c;
	header[DRV_TOP + 0x042c + 1] = 0xe0;
	header[DRV_TOP + 0x042c + 2] = 0xe1;
	header[DRV_TOP + 0x0508 + 0] = 0x4c;
	header[DRV_TOP + 0x0508 + 1] = 0xe0;
	header[DRV_TOP + 0x0508 + 2] = 0xe1;

	header[DRV_TOP + 0x055a + 0] = 0x68;
	header[DRV_TOP + 0x055a + 1] = 0x20;
	header[DRV_TOP + 0x055a + 2] = 0xf8;
	header[DRV_TOP + 0x055a + 3] = 0xe1;

	header[DRV_TOP + 0x0ff0] = 0xa9;
	/* header[DRV_TOP + 0x0ff1] = 0x??; */
	header[DRV_TOP + 0x0ff2] = 0x53;
	header[DRV_TOP + 0x0ff3] = 0x04;
	header[DRV_TOP + 0x0ff4] = 0x1a;
	header[DRV_TOP + 0x0ff5] = 0x53;
	header[DRV_TOP + 0x0ff6] = 0x08;
	header[DRV_TOP + 0x0ff7] = 0x73;
	/* header[DRV_TOP + 0x0ff8] = 0x??; */
	/* header[DRV_TOP + 0x0ff9] = 0x??; */
	header[DRV_TOP + 0x0ffa] = 0x00;
	header[DRV_TOP + 0x0ffb] = 0xd0;
	header[DRV_TOP + 0x0ffc] = 0x00;
	header[DRV_TOP + 0x0ffd] = 0x10;
	header[DRV_TOP + 0x0ffe] = 0x60;

/*
	AA
	BD 00 E2		lda	$E200,x		;
	8D F1 CF		sta $CFF1
	BD 00 E3		lda	$E300,x		;
	8D F8 CF		sta $CFF8
	BD 00 E4		lda	$E400,x		;
	8D F9 CF		sta $CFF9
	20 F0 CF
*/

	fwrite(&header[0x0],1,0x2000,ofp);
	op = 0x2000;
	song = 0;

	for (ip = 0; searchsong(ifp, &ip); ip += 1)
	{
		int ol;
		unsigned char inbuf[MUS_MAX];
		unsigned char outbuf[MUS_MAX];
		fseek(ifp, ip, SEEK_SET);
		il = fread(inbuf, 1, MUS_MAX, ifp);
		if (il < 16) break;

		ol = decode_ds8music(outbuf, MUS_MAX, inbuf, il);
		if (ol < 0)
		{
			fprintf(stderr, "decode error %08x:%s.\n", ip, errstr[ol + 64]);
			continue;
		}
		fprintf(stderr, "decode %08x:ok.\n", ip);
		l = op & 0xff;
		h = (op >> 8) & 0x1f;
		b = op >> 13;
		header[0x200 + song] = b;
		header[0x300 + song] = l;
		header[0x400 + song] = h + 0x40;
		fwrite(outbuf,1,ol,ofp);
		op += ol;
		song++;
	}
	if (op & 0x1fff)
	{
		unsigned char zerobuf[0x2000];
		int l = 0x2000 - (op & 0x1fff);
		memset(zerobuf,0,l);
		fwrite(zerobuf,1,l,ofp);
		op += l;
	}
	op -= 0x20;
	header[0x14] = (op >> (8 * 0)) & 255;
	header[0x15] = (op >> (8 * 1)) & 255;
	header[0x16] = (op >> (8 * 2)) & 255;
	header[0x17] = (op >> (8 * 3)) & 255;

	fseek(ofp, 0, SEEK_SET);
	fwrite(header,1,0x2000,ofp);
	fclose(ofp);
	fclose(ifp);
	fprintf(stderr, "%d songs.\n", song);
	return 0;
}
